%{

  /*! \file parser.c
    \brief svnaccess file parser.
  */

  /*
   * File version information:
   * $Id: parser.y 1640 2010-05-21 15:49:54Z corti $
   * $Revision: 1640 $
   * $HeadURL: https://svn.id.ethz.ch/bd_webhosting/tools/svnauthcheck/parser.y $
   * $Date: 2010-05-21 17:49:54 +0200 (Fri, 21 May 2010) $
   */

  
#include "svnauthcheck.h"
#include <stdio.h>
    
  char                            location[MAX_FILE_NAME_LENGTH];
  extern FILE                   * htaccess;
  extern char                     subtree[MAX_FILE_NAME_LENGTH];
  extern char                     user[MAX_USER_SIZE];
  extern char                   * yytext;
  extern int                      apache_flag;
#ifdef DEBUG
  extern int                      debug_flag;
#endif  
  extern int                      subtree_flag;
  extern int                      allow_ending_slashes_flag;
  extern int                      yylineno;
  extern char                     filename[MAX_FILE_NAME_LENGTH];
  extern struct user_type       * additional_readers;
  extern struct user_type       * anybody;
  extern struct user_type       * revoked_readers;
  extern struct user_type       * first_admin;
  int                             star;
  struct node                   * dir_node;
  struct user_type              * cur;
  struct user_type              * tmp_user;

  extern struct symbol          * table;

#ifdef TRAC_SUPPORT
  extern struct trac_permission * trac;
#endif

  struct user_type        * group;


  
  int  yylex        (void);
  void yyerror      (char const * message);
  void init_readers (void);
  
  /* initialize the reader lists */
  void init_readers() {
    additional_readers = NULL;
    revoked_readers    = NULL;
  }
 
  %}

%error-verbose

%union{
  struct user_type *list;
  char *string;
}

/*
 * Grammar (EBNF):
 * 
 * newlines       := { NEWLINE } .
 * file           := DirRuleList { TracPermission } .
 * group_def      := ID EQUAL IdList newlines { ID EQUAL IdList newlines }.
 * IdList         := id { COMMA id } .
 * id             := [ AT ] ID .
 * DirRuleList    := DirRule { DirRule } .
 * DirDefinition  := LBRACKET DIR RBRACKET newlines .
 * DirRule        := DirDefinition RuleList
 *                 | GROUPS newlines GroupDef
 *                 | LBRACKET [ ID SEMICOLON ] DIR RBRACKET newlines
 *                 | LBRACKET ID SEMICOLON DIR RBRACKET newlines RuleList .
 * RuleList       := rule { rule } .
 * rule           := ( id | STAR) EQUAL [ ID ] newlines .
 * TracPermission := TRAC ID EQUAL ID newlines .
 *
 */

%token GROUPS SEMICOLON LBRACKET RBRACKET EQUAL AT STAR COMMA NEWLINE TRAC

%token <string> ID
%token <string> DIR

%type  <list>   id_list
%type  <list>   id
%type  <list>   dir_rule
%type  <list>   dir_rule_list
%type  <list>   file
%type  <list>   rule
%type  <list>   rule_list

%start file

%%

 /* this is necessary since two newlines could be separated by a comment
  * line and thus seen as two distinct tokens by the scanner */
newlines : NEWLINE
| NEWLINE newlines
;


/******************************************************************************
 * file  := DirRuleList { TracPermission } .
 ******************************************************************************/
file : /* empty */
| dir_rule_list {
  first_admin = $1;
}
| dir_rule_list trac_permissions_list {
  first_admin = $1;
}
;

/******************************************************************************
 * group_def := ID EQUAL IdList newlines { ID EQUAL IdList newlines }.
 ******************************************************************************/
group_def : ID EQUAL id_list newlines {
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: group_def: ID(%s) = id_list(", yylineno, $1);
    print_list($3);
    printf(")\n");
  }
#endif
  create_group($1, $3);
  if ($1) { free($1); }
}           
| group_def ID EQUAL id_list newlines {
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: group_def: ID(%s) = id_list(", yylineno, $2);
    print_list($4);
    printf(")\n");
  }
#endif
  create_group($2, $4);
  if ($2) { free($2); }
}
;

/******************************************************************************
 * IdList         := id { COMMA id } .
 ******************************************************************************/
id_list : id {
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: id_list : id(", yylineno);
    print_list($1);
    printf(")\n");
  }
#endif
  $$ = $1;
}
| id_list COMMA id {
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: id_list : id_list(", yylineno);
    print_list($1);
    printf(" COMMA id(");
    print_list($3);
    printf(")\n");
  }
#endif
  $$ = append($1, $3);
}
;

/******************************************************************************
 * id := [ AT ] ID .
 ******************************************************************************/
id : ID {
#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: id : ID(%s)\n", yylineno, $1); }
#endif
  $$ = new_user($1, 0);
  if ($1) { free($1); }
}
| AT ID {
#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: id := @ %s\n", yylineno, $2); }
#endif
  
  /* we can check groups only for complete files */
  if (!subtree_flag) {
    
    /* get the group from the symbol table */
    group = get_group($2);
    
    if (group == NULL) {

      /* we put the group in the list and substitute later when all the file is parsed */
      group = new_user($2, 1);
            
    }
    
    $$ = clone(group);
    
  } else {
    
    /* we are not able to check the group we just return a dummy user */
    $$ = new_user($2, 0);
    
  }
  
  if ($2) { free($2); }
  
}

;

/******************************************************************************
 * DirRuleList    := DirRule { DirRule } .
 ******************************************************************************/
dir_rule_list : dir_rule {
#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: dir_rule_list := dir_rule\n", yylineno); }
#endif
  $$ = $1;
}
| dir_rule_list dir_rule {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: dir_rule_list := dir_rule_list dir_rule\n", yylineno); }
#endif

  if ($2 != NULL) {
    $1 = append($1, clone($2));
  }

  $$ = $1;		  

}
;

/******************************************************************************
 * DirRule        := DirDefinition RuleList
 *                 | GROUPS newlines GroupDef
 *                 | LBRACKET [ ID SEMICOLON ] DIR RBRACKET newlines
 *                 | LBRACKET ID SEMICOLON DIR RBRACKET newlines RuleList .
 ******************************************************************************/
dir_rule : GROUPS {

#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: dir_rule := GROUPS newlines group_def\n", yylineno);
  }
#endif

} newlines group_def {

  $$ = NULL;

}

| LBRACKET DIR RBRACKET newlines { init_readers(); } rule_list {

  if (apache_flag) {
    strncpy(location, $2, MAX_FILE_NAME_LENGTH);
    star = FALSE;
  }

  if (! allow_ending_slashes_flag) {    
    check_ending_slash($2);
  }
  
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: dir_rule := [ DIR ] rule_list\n", yylineno);
  }
#endif
  
  if (subtree_flag) {
#ifdef HAVE_STRNLEN
    int subtree_len = strnlen(subtree, MAX_USER_SIZE);
    int dir_len     = strnlen($2,      MAX_USER_SIZE);
#else
    int subtree_len = strlen(subtree);
    int dir_len     = strlen($2);
#endif
    if (strncmp($2, subtree, subtree_len) == 0) {
      if ((dir_len > subtree_len) &&
          ($2[subtree_len] != '/')) {
        fprintf(stderr, "Error: directory %s is not in the current subtree\n", $2);
        exit(EXIT_FAILURE);
      }    
    } else {
      fprintf(stderr, "Error: directory %s is not in the current subtree\n", $2);
      exit(EXIT_FAILURE);
    }
  }

  $$ = NULL;
  if (subtree_flag) {
    if (strcmp($2, subtree) == 0) {
      $$ = $6;
    }
  } else {
    if (strcmp($2,"/") == 0) {
      $$ = $6;
    }
  }
  
  if (apache_flag) {

    dir_node = malloc(sizeof(struct node));
    dir_node->name  = strdup($2);
    dir_node->readers = additional_readers;
    dir_node->revoked = revoked_readers;
    dir_node->parent  = NULL;

    add_node("", dir_node);
  }

  if ($2) { free($2); }

}

/*         $2           $4           $6       $7        */
| LBRACKET ID SEMICOLON DIR RBRACKET newlines { init_readers(); } rule_list {

  if (apache_flag) {
    strncpy(location, $4, MAX_FILE_NAME_LENGTH);
    star = FALSE;
  }
#ifdef DEBUG
  if (debug_flag > 2) {
    printf("[debug-grammar] %i: dir_rule := [ %s : %s ]\n", yylineno, $2, $4);
  }
#endif

  if (subtree_flag) {
#ifdef HAVE_STRNLEN
    int subtree_len = strnlen(subtree, MAX_USER_SIZE);
    int dir_len     = strnlen($4,      MAX_USER_SIZE);
#else
    int subtree_len = strlen(subtree);
    int dir_len     = strlen($4);
#endif
    if (strncmp($4, subtree, subtree_len) == 0) {
      if ((dir_len > subtree_len) &&
          ($4[subtree_len] != '/')) {
        fprintf(stderr, "Error: directory %s is not in the current subtree\n", $4);
        exit(EXIT_FAILURE);
      }    
    } else {
      fprintf(stderr, "Error: directory %s is not in the current subtree\n", $4);
      exit(EXIT_FAILURE);
    }
  }
  
  $$ = NULL;
  if (subtree_flag) {
    if (strcmp($4, subtree) == 0) {
      $$ = $8;
    }
  } else {
    if (strcmp($4,"/") == 0) {
      $$ = $8;
    }
  }

  if (apache_flag) {

    dir_node = malloc(sizeof(struct node));
    dir_node->name  = strdup($4);
    dir_node->readers = additional_readers;
    dir_node->revoked = revoked_readers;
    dir_node->parent  = NULL;

    add_node($2, dir_node);
  }

  if ($2) { free($2); }
  if ($4) { free($4); }

}
| LBRACKET DIR RBRACKET newlines {
#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: dir_rule := [ %s ]\n", yylineno, $2); }
#endif
  if ($2) { free($2); }
  $$ = NULL;
}
| LBRACKET ID SEMICOLON DIR RBRACKET newlines {
  if ($2) { free($2); }
  if ($4) { free($4); }
  $$ = NULL;
}
;

/******************************************************************************
 * RuleList       := rule { rule } .
 ******************************************************************************/
rule_list : rule {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: rule : rule\n", yylineno); }
#endif

  $$ = $1;

}
| rule rule_list {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: rule : rule rule_list\n", yylineno); }
#endif

  $1 = append($1, clone($2));
  $$ = $1;

}
;

/* $1       $3          */
rule    : id EQUAL ID newlines {            

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: ID (%s) = ID (%s) \\n\n", yylineno, $1->name, $3); }
#endif

  cur = $1;
  while(cur) {

    if (!contains(additional_readers, cur)) {
      additional_readers = append(additional_readers, new_user(cur->name, cur->unexpanded_group));
    }

    cur = cur->next;

  }

  if (strcmp($3, "rw") == 0) {
    $$ = $1;
  } else {
    $$ = NULL;
  }

  if ($3) { free($3); }

}
| id EQUAL newlines {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: ID (%s) = \\n\n", yylineno, $1->name); }
#endif

  cur = $1;
  while(cur) {

    if (!contains(revoked_readers, cur)) {
      revoked_readers = append(revoked_readers, new_user(cur->name, cur->unexpanded_group));
      cur = cur->next;
    }
  }

}
| STAR EQUAL ID newlines {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: * = \\n\n", yylineno); }
#endif

  star = TRUE;

  if (strcmp($3, "rw") == 0) {
    /* everybody is a writer */	     
  }

  if (!contains(additional_readers, anybody->name)) {
    additional_readers = append(additional_readers, clone(anybody));
  }

  if ($3) { free($3); }

  $$ = NULL;

}

| STAR EQUAL newlines {

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: * = \\n\n", yylineno); }
#endif

  if (!contains(additional_readers, anybody)) {
    revoked_readers = append(revoked_readers, clone(anybody));
  }

  $$ = NULL;

}

;

/******************************************************************************
 * TracPermission := TRAC ID EQUAL ID newlines .
 ******************************************************************************/
trac_permissions_list : trac_permissions
                      | trac_permissions_list trac_permissions
;

trac_permissions      : TRAC ID EQUAL ID newlines {

#ifdef TRAC_SUPPORT

  struct trac_permission * new_permission;

#ifdef DEBUG
  if (debug_flag > 2) { printf("[debug-grammar] %i: TRAC %s = %s\n", yylineno, $2, $4); }
#endif

  /* append the new permission information to the trac permissions list */
  
  new_permission = malloc(sizeof(struct trac_permission));
  if (new_permission == NULL) {
    fprintf(stderr, "Error: out of memory\n");
    exit(EXIT_FAILURE);
  }

  new_permission->user = malloc(strlen($4)+1);
  if (new_permission->user == NULL) {
    fprintf(stderr, "Error: out of memory\n");
    exit(EXIT_FAILURE);
  }
  strcpy(new_permission->user, $4);
  
  new_permission->permission = malloc(strlen($2)+1);
  if (new_permission->permission == NULL) {
    fprintf(stderr, "Error: out of memory\n");
    exit(EXIT_FAILURE);
  }
  strcpy(new_permission->permission, $2);

  new_permission->next = trac;
  trac                 = new_permission;
  
  if ($2) { free($2); }
  if ($4) { free($4); }

#endif

}

;

%%

void yyerror (char const * message) {
  fprintf(stderr, "%s:%i: %s (%s)\n", filename, yylineno, message, yytext);
  exit(SYNTAX_ERROR);
}


