/*! \file svnauthcheck.c
    \brief Main program.

    Main program.
*/

/*
 * File version information:
 * $Id: svnauthcheck.c 1636 2010-05-20 11:08:34Z corti $
 * $Revision: 1636 $
 * $HeadURL: https://svn.id.ethz.ch/bd_webhosting/tools/svnauthcheck/svnauthcheck.c $
 * $Date: 2010-05-20 13:08:34 +0200 (Thu, 20 May 2010) $
 */

#include "svnauthcheck.h"

extern FILE * yyin; /*!< File to be scanned and parsed */

/* command line arguments */
char                       filename[MAX_FILE_NAME_LENGTH];      /*!< name of the input file */
char                       subtree[MAX_FILE_NAME_LENGTH];       /*!< Subtree being checked */
int                        allow_ending_slashes_flag = FALSE;   /*!< Command line argument: true if ending slashed are allowed */
int                        apache_flag   = FALSE;               /*!< Command line argument: generate Apache permissions */
int                        list_flag     = FALSE;               /*!< Command line argument: list admins */
int                        trac_flag     = FALSE;               /*!< Command line argument: modify trac permissions */
#ifdef DEBUG
int                        debug_flag    = 0;                   /*!< Command line argument: debugging level */
#endif
int                        subtree_flag  = FALSE;               /*!< Command line argument: true if we are checking only part of the tree */
#ifdef TRAC_SUPPORT
int                        count_result  = 0;                   /*!<  The result of the count sqlite callback routine */
#endif
static int                 template_flag = FALSE;               /*!< Command line argument: true if a configuration template was specified */
static int                 user_flag    = FALSE;                /*!< Command line argument: true if a user was specified */
static int                 verbose_flag = 0;                    /*!< Command line argument: enable verbose output */
static int                 version_flag = FALSE;                /*!< Command line argument: true if the version has to printed */

static FILE              * htaccess;                            /*!< htaccess file */
static char                htaccess_name[MAX_FILE_NAME_LENGTH]; /*!< Name of the htaccess file */
static char                path[MAX_FILE_NAME_LENGTH];          /*!< Path of the current location */
static char                trac_dir[MAX_FILE_NAME_LENGTH] ;     /*!< Path of the trac project */
static char                prefix[MAX_FILE_NAME_LENGTH];        /*!< Prefix for all locations */
static char                template[MAX_FILE_NAME_LENGTH];      /*!< Name of the Apache permission template file */
static char                user[MAX_USER_SIZE];                 /*!< Command line argument that specifies a user that has to be able to write */
static char              * template_buf;                        /*!< Memory buffer containing the Apache permission template */
static struct repository * repositories = NULL;                 /*!< The list of repositories */
struct user_type         * additional_readers;                  /*!< Users allowed to read the given location */
struct user_type         * anybody;                             /*!< Special user representing the "*" in svnaccess and "valid-user" in Apache */
struct user_type         * first_admin = NULL;                  /*!< A special user that has to have write access */
struct user_type         * revoked_readers;                     /*!< Users which read rights where revoked for the given directory */

struct symbol            * table;                               /*!< User group table: the table contains only user groups and not
                                                                 *   single users */
#ifdef TRAC_SUPPORT
struct trac_permission   * trac;                                /*!< List of optional trac permissions */
#endif

/******************************************************************************
 * Symbol table
 ******************************************************************************/

/*!
 * \brief          Retrieves a group from the user table
 * \param[in] name Group name
 * \return         Group members
 */
struct user_type * get_group(char * name) {

  struct symbol * cur;

#ifdef DEBUG
  if (debug_flag > 1) {
    printf("[debug] get_group(%s) ", name);
  }
#endif

  cur = table;
  while (cur) {
    if (strncmp(cur->name, name, MAX_USER_SIZE) == 0) {
      /* the symbol is already in the table */
#ifdef DEBUG
      if (debug_flag > 1) {
        printf("[debug] get_group => ");
        print_list(cur->list);
        printf("\n");
      }
#endif
      return cur->list;
    }
    cur = cur->next;
  }

#ifdef DEBUG
      if (debug_flag > 1) {
        printf("[debug] get_group => NULL\n");
      }
#endif

  return NULL;

}

/*!
 * \brief          Generates an error if the specified path ends in a /
 * \param[in] path Path to check
 */
void check_ending_slash(char * directory_path) {

#ifdef HAVE_STRNLEN
  int dir_len = strnlen(directory_path,      MAX_USER_SIZE);
#else
  int dir_len = strlen(directory_path);
#endif

  if (dir_len > 1) {

    if ( directory_path[dir_len-1] == '/' ) {
      fprintf(stderr, "Error: ending slash for '%s'.\n", directory_path);
      exit(EXIT_FAILURE);
    }

  } else {

    if (directory_path[0] != '/') {
      fprintf(stderr, "Error: wrong repository root '%s'.\n", directory_path);
      exit(EXIT_FAILURE);
    }

  }
}

/*!
 * \brief          checks if a character is a valid hexadecimal digit
 * \param[in] c    digit to check
 * \return         true if valid
 */
int valid_digit(char c) {

  return
    ( (c >= 'a') && (c <= 'z') ) ||
    ( (c >= 'A') && (c <= 'A') ) ||
    ( (c >= '0') && (c <= '9') );

}

/*!
 * \brief            URL-decodes (in place) a string
 * \param[in] string pointer to the string to decode
 */
void urldecode( char * string ) {

  char * write_pos; /* point we are overwriting */
  char * read_pos;  /* point we arr reading from: will be write pos +2 every time we decode */

  char code[3];

  read_pos  = string;
  write_pos = string;

  while ( * read_pos != '\0' ) {

    if ( * read_pos == '%' ) {

      /* check if the string terminates before the end of the code */
      if ( ( *(read_pos+1) == '\0' ) ||
           ( *(read_pos+2) == '\0' ) ||
           ! ( valid_digit( *(read_pos+1) ) ) ||
           ! ( valid_digit( *(read_pos+2) ) ) ) {
        fprintf(stderr, "Error: wrong URL encoding in %s.\n", string);
        exit(EXIT_FAILURE);
      }

      /* decode the next two chars*/
      code[0] = *(read_pos+1);
      code[1] = *(read_pos+2);
      code[2] = '\0';

      *write_pos = (char)strtoul(code, NULL, 16);

      read_pos += 3;

    } else {

      *write_pos = *read_pos;
      read_pos++;

    }

    write_pos++;

  }

  *write_pos = '\0';

}

/**
 * @brief       Create a new group of users
 * @param  name Name of the group
 * @return      The entry in the symbol table
 */
struct symbol * create_group(char * name, struct user_type * users) {

  struct symbol * sym;

  /* sanity check: the user should no be in the list */
  if (get_group(name) != NULL) {
    fprintf(stderr, "Error: Duplicate group %s encountered.\n", name);
    exit(EXIT_FAILURE);
  }

#ifdef DEBUG
  if (debug_flag > 1) {
    printf("[debug] create_group(%s, ", name);
    print_list(users);
    printf(")\n");
  }
#endif

  /* allocate */
  sym = malloc(sizeof(struct symbol));
  if (sym == NULL) error("Out of memory");

  /* fill the fields */
  strncpy(sym->name, name, MAX_USER_SIZE);
  sym->list = users;

  /* attach the symbol to the symbol table */
  sym->next = table;
  table     = sym;

  return sym;

}

/**
 * @brief Frees the storage used by the groups symbol table
 */
void free_groups() {

  struct symbol * cur = table;
  struct symbol * tmp;

  while(cur) {
    free_users(cur->list);
    tmp = cur;
    cur = cur->next;
    free(tmp);
  }

}

/******************************************************************************
 * User management
 ******************************************************************************/

/**
 * @brief      Creates a new user structure
 * @param name The user name
 * @return     The newly created structure
 */
struct user_type * new_user(char * name, int unexpanded_group) {

  struct user_type * usr;

  /* sanity check on the name length */
#ifdef HAVE_STRNLEN
  if (strnlen(name, MAX_USER_SIZE) == MAX_USER_SIZE) {
#else
  if (strlen(name) == MAX_USER_SIZE) {
#endif
    fprintf(stderr, "Error: %s is too long\n", name);
    exit(EXIT_FAILURE);
  }

  usr = malloc(sizeof(struct user_type));
  if (usr == NULL) error("Out of memory");

  strncpy(usr->name, name, MAX_USER_SIZE);
  usr->unexpanded_group = unexpanded_group; /* default users are users not groups */
  usr->next = NULL;
  usr->prev = NULL;

  return usr;

}

/**
 * @brief      Frees storage uses by a user list
 * @param list User list
 */
void free_users(struct user_type * list) {

  struct user_type * cur = list;
  struct user_type * tmp;

  /* sanity checks */
  if (list == NULL) {
    return;
  }

  while (cur) {
    tmp = cur;
    cur = cur->next;
    free(tmp);
  }

}

/**
 * @brief     Checks if a user is already in a given list
 * @param src The list
 * @param usr The user
 * @return    1 (true) if the list contains the user
 */
int contains(struct user_type * src, struct user_type * usr) {

  struct user_type * cur;

#ifdef DEBUG
  if (debug_flag > 1) {
    printf("[debug] contains(");
    print_list(src);
    printf(", %s)\n", usr->name);
  }
#endif

  /* sanity checks */
  if (src == NULL) {
#ifdef DEBUG
    if (debug_flag > 1) {
      printf("[debug] contains => FALSE\n");
    }
#endif
    return FALSE;
  }

  assert(usr != NULL);

  cur = src;
  while (cur) {
    if (strncmp(cur->name, usr->name, MAX_USER_SIZE) == 0) {
#ifdef DEBUG
      if (debug_flag > 1) {
        printf("[debug] contains => TRUE\n");
      }
#endif
      return TRUE;
    }
    cur = cur->next;
  }

#ifdef DEBUG
  if (debug_flag > 1) {
    printf("[debug] contains => FALSE\n");
  }
#endif
  return FALSE;

}

/**
 * @brief      Appends a user list to another removing duplicates
 * @param dest The destination list
 * @param src  The list to be appended
 * @return     The merged list (without duplicates)
 */
struct user_type * append(struct user_type * dest, struct user_type * src) {

  struct user_type * cur_dest;
  struct user_type * cur_src;
  struct user_type * tmp;

#ifdef DEBUG
  if (debug_flag > 1) {
    printf("[debug] append(");
    print_list(dest);
    printf(", ");
    print_list(src);
    printf(")\n");
  }
#endif

  if (dest == NULL) {
#ifdef DEBUG
    if (debug_flag > 1) {
      printf("[debug] append => ");
      print_list(src);
      printf("\n");
    }
#endif
    return src;
  }

  if (src == NULL) {
    return dest;
  }

  /* go to the end of the dest list */
  cur_dest = dest;
  while (cur_dest->next) {
    cur_dest = cur_dest->next;
  }

  /* cur_dest is now the last element of dest */
  assert(cur_dest->next == NULL);

  /* eliminate duplicates */
  cur_src = src;
  while (cur_src) {
    tmp     = cur_src;
    cur_src = cur_src->next;
    if (contains(dest, tmp)) {
      /* remove it */

      if (tmp->prev != NULL) { tmp->prev->next = tmp->next; }
      if (tmp->next != NULL) { tmp->next->prev = tmp->prev; }

      if (tmp == src) {
        /* we are eliminating the first element */
        src = tmp->next;
      }

      free(tmp);
    }
  }

  if (src != NULL) {
    /* some new elements are available */
    cur_dest->next = src;
    src->prev      = cur_dest;
  }

#ifdef DEBUG
    if (debug_flag > 1) {
      printf("[debug] append => ");
      print_list(dest);
      printf("\n");
    }
#endif
  return dest;

}

/**
 * @brief      Clones a user lisr
 * @param list The list to be cloned
 * @return     A clone (shallow copy) of the given list
 */
struct user_type * clone(struct user_type * list) {

  struct user_type * cur        = NULL;
  struct user_type * new_list   = NULL;
  struct user_type * user_clone = NULL;

  if (list == NULL) {
    return NULL;
  }

  cur = list;
  while (cur) {

    user_clone = new_user(cur->name, cur->unexpanded_group);
    
    user_clone->next = new_list;
    if (new_list != NULL) {
      new_list->prev = user_clone;
    }
    new_list         = user_clone;

    cur = cur->next;

  }

  return new_list;

}

/**
 * @brief      Prints a user list on stdout (comma separated)
 * @param list The list to print
 */
void print_list(struct user_type * list) {

  struct user_type * cur;

  cur = list;
  if (cur) {
    printf("%s", cur->name);
    cur = cur->next;
    while(cur) {
      printf(",%s", cur->name);
      cur = cur->next;
    }
  } else {
    printf("NULL");
  }

}

/******************************************************************************
 * Helper functions
 ******************************************************************************/

/*!
 * \brief Prints usage information
 */
void prog_usage() {

  printf("\nusage: [OPTIONS] [AUTHFILE]\n\n");
  printf("Options\n");
  printf(" -a, --apache=file           generates an apache-style permission file to be used with ViewCV\n");
  printf("     --allow-ending-slashes  allows a / at the end of directories\n");
  printf(" -d, --debug                 enables debugging output\n");
  printf(" -h, --help                  this help message\n");
  printf(" -l, --list-admins           lists repository administrators\n");
  printf(" -p, --prefix=dir            prefix to prepend to the apache-style permissions\n");
  printf(" -s, --subtree=dir           the checked file is a partial svnaccess file for the given subtree\n");
  printf(" -t, --template=file         a template to be used to generate the apache style permission file\n");
#ifdef TRAC_SUPPORT
  printf(" -T, --trac=path             the path of the trac project dir\n");
#endif
  printf(" -u, --user=user             specifies a user that has to be able to write at the repository root\n");
  printf(" -v, --verbose               increase verbosity\n");
  printf(" -V, --version               prints the version of svncheck\n");
  printf("\n\nPlease see the svnauthcheck(1) man page for full documentation\n");

}

/*!
 * \brief Prints and error message and exits
 * \param[in] message Error message
 */
void error(const char * message) {
  fprintf(stderr, "Error: %s\n", message);
  exit(EXIT_FAILURE);
}

/******************************************************************************
 * htaccess
 ******************************************************************************/

/*!
 * \brief          Reads an Apache permission template to a memory buffer
 * \param[in] name Name of the file containing the template
 * \return         Address of the memory buffer containg the template
 */
char * read_template(char * name) {

  FILE   *f;
  char   *buf;
  long    len;
  size_t  tmp;

  char    msg[MAX_MSG_LEN];

  if (!(f = fopen(name, "r"))) {
    strncpy(msg, "cannot open template file ", MAX_MSG_LEN);
    strncat(msg, name, MAX_MSG_LEN);
    error(msg);
  }

  /* allocate buffer */
  if (fseek(f,0, SEEK_END) < 0) {
    strncpy(msg, "seeking at the end of the template file ", MAX_MSG_LEN);
    strncat(msg, name, MAX_MSG_LEN);
    error(msg);
  }
  len = ftell(f);

#ifdef DEBUG
  if (debug_flag > 0) {
    printf("[debug] opening template \"%s\": %li bytes\n", name, len);
  }
#endif

  buf = malloc((size_t)len);

  if (fseek(f,0, SEEK_SET) < 0) {
    strncpy(msg, "seeking at the beginning of the template file ", MAX_MSG_LEN);
    strncat(msg, name, MAX_MSG_LEN);
    error(msg);
  }

  if ((tmp = fread(buf, (size_t)len, sizeof(char), f)) != 1) {
    strncpy(msg, "problem reading ", MAX_MSG_LEN);
    strncat(msg, name, MAX_MSG_LEN);
    strncat(msg, " (", MAX_MSG_LEN);
    strncat(msg, strerror(errno), MAX_MSG_LEN);
    strncat(msg, ")", MAX_MSG_LEN);
    error(msg);
  }

  if (fclose(f) == EOF) {
    strncpy(msg, "cannot close the template file ", MAX_MSG_LEN);
    strncat(msg, name, MAX_MSG_LEN);
    error(msg);
  }

  return buf;

}

/*!
 * \brief Writes a &lt;Location&gt; directive for the given directory
 * \param[in] f        file to write to
 * \param[in] location location string
 * \param[in] n        corresponding directory tree node
 */
void write_location (FILE * f, char * location, struct node * n) {

  struct user_type * cur;

  fprintf(f, "<Location %s%s>\n", prefix, location);

  // \TODO Mmmm this in principle is not safe since template_buf could be anything
  fprintf(f, "%s", template_buf);

  if (contains(n->users, anybody)) {

    fprintf(f, "Require valid-user\n");

  } else {

    fprintf(f, "Require user");

    cur = n->users;
    while (cur != NULL) {
      fprintf(f, " %s", cur->name);
      cur = cur->next;
    }

    fprintf(f, "\n");

  }

  fprintf(f, "</Location>\n");

}

/******************************************************************************
 * Repositories
 ******************************************************************************/

/*!
 * \brief Returns the data structure for the given repository (if it does not
 *        exist it is instantiated)
 * \param[in] name The repository name
 */
 struct repository * get_repository(const char * name) {

  struct repository * cur = repositories;
  struct repository * new_repo;

#ifdef DEBUG
  if (debug_flag) { printf("[debug] get_repository(%s)\n", name); }
#endif

  while (cur) {

    if (strncmp(cur->name, name, MAX_REPOSITORY_SIZE) == 0) {
      /* we found the matching repository */
      return cur;
    }

    cur = cur->next;

  }

  /* instantiate a new repository */
  new_repo = malloc(sizeof(struct repository));
  if (new_repo == NULL) {
    out_of_memory();
  }

  /* initialize fields */
  new_repo->root = NULL;
  new_repo->next = NULL;
#ifdef HAVE_STRNDUP
  new_repo->name = strndup(name, MAX_REPOSITORY_SIZE);
#else
  new_repo->name = strdup(name);
#endif

  init_dir_tree(new_repo);

  /* attach it to the repository list */
  new_repo->next = repositories;
  repositories   = new_repo;

  return new_repo;

}

/******************************************************************************
 * Dir tree
 ******************************************************************************/

/*!
 * \brief Initialize the directory tree
 * \param[in] repo The repository to initialize
 */
void init_dir_tree(struct repository * repo) {

#ifdef DEBUG
  if (debug_flag > 0) {
    printf("[debug] initializing the directory tree for %s\n",
           repo->name);
  }
#endif

  repo->root = malloc(sizeof(struct node));
  repo->root->name     = strdup("/");
  repo->root->readers  = anybody; /* default */
  repo->root->revoked  = NULL;
  repo->root->children = NULL;
  repo->root->parent   = NULL;
  repo->root->users    = NULL;

}

/*!
 * \brief          Searches for a node among the children of the given node
 * \param[in] node Subtree to be searched
 * \param[in] name Node name
 * \return         Searched node
 *
 * Search a node with name "name" among the children of the given. If the node is
 * not found a new node is created.
 */
struct node * dir_search(struct node * node, char * name) {

  struct child * c    = node->children;
  struct child * last = c;
  struct node  * n;

#ifdef DEBUG
  if (debug_flag > 0) {
    printf("[debug] looking for %s among the children of %s\n", name, node->name);
  }
#endif

  while (c != NULL && strcmp(c->node->name, name) != 0) {
    last = c;
    c = c->next;
  }

  if (c == NULL) {

    /* node not found */

    n = malloc(sizeof(struct node));
    if (n == NULL) {
      out_of_memory();
    }

    n->name     = strdup(name);
    n->readers  = NULL;
    n->revoked  = NULL;
    n->children = NULL;
    n->users    = NULL;
    n->parent   = node;

#ifdef DEBUG
    if (debug_flag > 0) {
      printf("[debug] adding a new node %s\n", n->name);
    }
#endif

    c = malloc(sizeof(struct child));
    if (c == NULL) {
      out_of_memory();
    }

    c->node = n;
    c->next = NULL;

    if (node->children == NULL) {
      node->children = c;
    } else {
      last->next     = c;
    }

    return n;

  }

#ifdef DEBUG
  if (debug_flag > 0) {
    printf("[debug] found node %s\n", c->node->name);
  }
#endif

  return c->node;

}

/*!
 * \brief       Add a directory node to the directory tree
 * \param[in] repo_name   The name of the repository
 * \param[in] n           Node to add
 */
void add_node(const char * repo_name, struct node * n) {

  char              * cur_path;
  char              * dir_name;
  struct node       * node;
  struct repository * repo = get_repository(repo_name);

#ifdef DEBUG
  if (debug_flag > 0) {
    printf("[debug] adding node (repository: %s): %s +{",
           repo_name,
           n->name);
    print_list(n->readers);
    printf("} -{");
    print_list(n->revoked);
    printf("}\n");
  }
#endif

  if (strcmp(n->name, "/") == 0) {

    /* the root is a special case */
    repo->root->readers = n->readers;
    repo->root->revoked = n->revoked;

  } else {

    /* split name components */
    cur_path = strdup(n->name);

    node = repo->root;

    while((dir_name = strsep(&cur_path, "/")) != NULL) {

      if (strcmp(dir_name, "") != 0) {

#ifdef DEBUG
        if (debug_flag > 0) {
          printf("[debug]   dir: %s\n", dir_name);
        }
#endif

        node = dir_search(node, dir_name);

      }
    }

    if (node) {

      /* the node already exists */
      node->readers = append(node->readers, clone(n->readers));
      node->revoked = append(node->revoked, clone(n->revoked));

    }

    if (cur_path) {
      free(cur_path);
    }

  }

#ifdef DEBUG
  if (debug_flag) {
    debug_tree(repo->root, 0);
  }
#endif


}

#ifdef DEBUG
/*!
 * \brief Prints a directory tree node
 * \param[in] n tree node
 */
void print_node(struct node * n) {

  printf("%s (u: ", n->name);
  print_list(n->users);
  printf(", r: ");
  print_list(n->readers);
  printf(", rvk: ");
  print_list(n->revoked);
  printf(")");

}

/*!
 * \brief Prints a directory tree
 * \param[in] n     tree root
 * \param[in] level initial indentation
 */
void debug_tree(struct node * n, int level) {

  int            i;
  struct child * cur;

  printf("[debug] tree: ");
  for (i = 0; i < level; i++) {
    printf(" ");
  }
  printf("%s +{", n->name);
  print_list(n->readers);
  printf("} -{");
  print_list(n->revoked);
  printf("}\n");

  cur = n->children;
  while (cur) {
    assert(cur->node->parent == n); /* check that the parent link was correctly set */
    debug_tree(cur->node, level+2);
    cur = cur->next;
  }

}
#endif

void check_groups(struct user_type * u) {

  struct user_type * cur;
  cur = u;
  while (cur) {

    if (cur->unexpanded_group &&          // the entry is a group
        get_group(cur->name) == NULL) {

      /* Unknown group */

#ifdef HAVE_STRNLEN
      char * msg = malloc(strlen("unknown group ") +
                          strnlen(cur->name, MAX_USER_SIZE));
#else
      char * msg = malloc(strlen("unknown group ") +
                          strlen(cur->name));
#endif
      strcpy(msg, "unknown group ");
      strcat(msg, cur->name);

      fprintf(stderr, "Error: %s\n", msg);
      exit(EXIT_FAILURE);

    }
    cur = cur->next;
  }

}

void check_repository_groups(struct node * n) {

  check_groups(n->readers);
  check_groups(n->users);
  check_groups(n->revoked);

}

/*!
 * \brief Converts a tree of directories to Apache locations
 * \param[in] n          Tree root
 * \param[in] full_path  The full path of the location
 * \todo                 We should generate a location only if it
 *                       differs from the parent
 */
void convert_tree(struct node * n, char * full_path) {

  struct child *     c;
  char               cur_path[MAX_FILE_NAME_LENGTH];
  struct user_type * cur;

  strcpy(cur_path, full_path);

  if (n->name[0] != '/' && cur_path[strlen(cur_path)-1] != '/') {
    strcat(cur_path, "/");
  }

  strcat(cur_path, n->name);

  /*
   * build the final access list
   *   users = (readers + parent->readers) - revoked
   *   users = (readers - revoked) + (parent->readers - revoked)
   */

  /* users = readers - revoked */
  cur = n->readers;
  while (cur) {
    if (!contains(n->revoked, cur)) {
      n->users = append(n->users, new_user(cur->name, cur->unexpanded_group));
    }
    cur = cur->next;
  }

  /* users += parent->readers - revoked */
  if (n->parent != NULL) {
    cur = n->parent->readers;
    while (cur) {
      if (!contains(n->revoked, cur) &&
          !contains(n->users,   cur)) {
        n->users = append(n->users, new_user(cur->name, cur->unexpanded_group));
      }
      cur = cur->next;
    }
  }

  write_location (htaccess, cur_path, n);

  c = n->children;
  while (c) {
    convert_tree(c->node, cur_path);
    c = c->next;
  }

}

#ifdef TRAC_SUPPORT
/******************************************************************************
 * sqlite callbacks
 ******************************************************************************/

static int count_callback(void  *pArg,
                          int    argc,
                          char **argv,
			  char **columnNames) {
  count_result = atoi(argv[0]);
  return 0;
}


#endif

/******************************************************************************
 * Main
 ******************************************************************************/

/*!
 * \brief          Main
 * \param[in] argc Number of command line arguments
 * \param[in] argv Command line arguments
 * \return         Exit status
 */
int main(int argc, char **argv) {

  /* getopt_long stores the option index here. */
  int option_index = 0;
  int c;

  static struct option long_options[] = {
    {"apache",               optional_argument, NULL,          'a'},
    {"allow-ending-slashes", optional_argument, NULL,          ALLOW_ENDING_SLASHES_OPT},
    {"debug",                no_argument,       NULL,          'd'},
    {"help",                 no_argument,       NULL,          'h'},
    {"list-admins",          no_argument,       NULL,          'l'},
    {"prefix",               required_argument, NULL,          'p'},
    {"subtree",              required_argument, NULL,          's'},
    {"template",             required_argument, NULL,          't'},
#ifdef TRAC_SUPPORT
    {"trac",                 required_argument, NULL,          'T'},
#endif
    {"user",                 required_argument, NULL,          'u'},
    {"verbose",              no_argument,       NULL,          'v'},
    {"version",              no_argument,       NULL,          'V'},
    {NULL,                   0,                 NULL,          0}
  }; /*!< Command line options */

  struct repository      * cur;

#ifdef TRAC_SUPPORT

  struct trac_permission * cur_trac;

#ifdef HAVE_SQLITE3_H
  sqlite3                * tracdb;
#else
  sqlite                 * tracdb;
#endif

#endif

  /* default values */
  strcpy(htaccess_name, ".htaccess");
  strcpy(prefix,        "");

  anybody = malloc(sizeof(struct user_type));
  if (anybody == NULL) {
    out_of_memory();
  }

  anybody->next = NULL;
  anybody->prev = NULL;
  strcpy(anybody->name, "anybody");

  while (TRUE) {

    c = getopt_long(argc,
                    argv,
#ifdef DEBUG
                    "a:t:hvu:s:dp:VT:",
#else
                    "a:t:hlvu:s:p:VT:",
#endif
                    long_options,
                    &option_index);

    /* Detect the end of the options. */
    if (c == -1)
      break;

    switch (c) {

    case 'a':
      apache_flag = TRUE;
      if (optarg) {
        strncpy(htaccess_name, optarg, MAX_FILE_NAME_LENGTH);
      } else {
        fprintf(stderr, "Error: --apache was specified without an argument\n");
        prog_usage();
        exit(EXIT_FAILURE);
      }
      break;

#ifdef DEBUG
    case 'd':
      debug_flag++;
      break;
#endif

    case 'h':
      prog_usage();
      exit(EXIT_SUCCESS);

    case 'l':
      list_flag = TRUE;
      break;

    case 'p':
      strncpy(prefix, optarg, MAX_FILE_NAME_LENGTH);
      break;

    case 's':
      if (optarg[0] != '/') {
        fprintf(stderr, "Error: subtree %s does not begin at the root (/) of the repository.\n", optarg);
        exit(EXIT_FAILURE);
      }
      strcpy(subtree,optarg);
      subtree_flag = TRUE;
      break;

    case 't':
      template_flag = TRUE;
      strncpy(template, optarg, MAX_FILE_NAME_LENGTH);
      break;

    case 'u':
      strncpy(user, optarg, MAX_USER_SIZE);
      user_flag = TRUE;
      break;

    case 'v':
      verbose_flag++;
      break;

    case 'V':
      version_flag = TRUE;
      break;

#ifdef TRAC_SUPPORT
    case 'T':
      strncpy(trac_dir, optarg, MAX_FILE_NAME_LENGTH);
      trac_flag = TRUE;
      break;
#endif

    case ALLOW_ENDING_SLASHES_OPT:
      allow_ending_slashes_flag = TRUE;
      break;

    case '?':
      prog_usage();
      /* getopt_long already printed an error message. */
      exit(USAGE_ERROR);

    default:
      abort ();
    }

  }

  if (apache_flag) {

    /* read template */

    if (template_flag) {
      template_buf = read_template(template);
    } else {
      /* empty string */
      template_buf    = malloc(1);
      template_buf[0] = '\0';
    }
    if (!(htaccess = fopen(htaccess_name, "w"))) {
      fprintf(stderr, "Error: cannot write to %s\n", htaccess_name);
      exit(EXIT_FAILURE);
    }

  }

  if (version_flag) {
    printf("svnauthcheck %s\n", PACKAGE_VERSION);
    exit(EXIT_SUCCESS);
  }

  if (verbose_flag) {
    printf("checking syntax\n");
  }

  /* parse */

  argc -= optind;
  argv += optind;

  if (argc > 0) {
    if (!(yyin = fopen(argv[0], "r"))) {
      fprintf(stderr, "Error: cannot open %s\n", argv[0]);
      exit(EXIT_FAILURE);
    }  else {
      strncpy(filename, argv[0], MAX_FILE_NAME_LENGTH);
      if (verbose_flag) {
        printf("parsing %s\n", argv[0]);
      }
    }
  } else {
    strncpy(filename, "-", MAX_FILE_NAME_LENGTH);
    yyin = stdin;
  }

  if (yyparse() != 0) {
    fprintf(stderr, "Error: unknown parser error.\n");
    exit(EXIT_FAILURE);
  }

  if (verbose_flag) {
    printf("%s parsed\n", argv[0]);
  }

  if (list_flag) {
    print_list(first_admin);
    printf("\n");
  }

  if (user_flag) {
    if (verbose_flag) {
      printf("Checking if user '%s' has permission to write.\n", user);
    }
    if (!contains(first_admin, new_user(user, 0))) {
      fprintf(stderr, "Error: user '%s' will not have write permission.\n", user);
      exit(EXIT_FAILURE);
    }
  }

  if(yyin != stdin) {
    fclose(yyin);
  }

  /*
   * Check undefined groups
   */

  cur = repositories;
  while(cur) {
    check_repository_groups(cur->root);
    cur = cur->next;
  }

  /*
   * Apache
   */

  if (apache_flag) {

    cur = repositories;

    while(cur) {

      strcpy(path, cur->name);
#ifdef DEBUG
      if (debug_flag) {
        printf("[debug] repository: %s\n", cur->name);
        debug_tree(cur->root, 0);
      }
#endif
      convert_tree(cur->root, path);

      cur = cur->next;

    }

  }

#ifdef TRAC_SUPPORT

  /*
   * Trac
   */

  if (trac_flag && trac != NULL) {

    char * sql_error;
    char * query;

    char * db_name = malloc(strlen(trac_dir)+strlen("/db/trac.db")+1);

    if (db_name == NULL) { error("Out of memory"); }
    strcpy(db_name, trac_dir);
    strcat(db_name, "/db/trac.db");

    if (verbose_flag) {
      printf("Opening %s\n", db_name);
    }
#ifdef HAVE_SQLITE3_H
    sqlite3_open(db_name, &tracdb);
#else
    tracdb = sqlite_open(db_name, 0, NULL);
#endif

    if (tracdb == NULL) {
      fprintf(stderr, "Error: Cannot open %s\n", db_name);
      exit(EXIT_FAILURE);
    }

    cur_trac = trac;
    while (cur_trac) {

      /* check if the permission is already in the DB */

      /* SELECT COUNT(*)
         FROM   permission
         WHERE  username = <user>
         AND    action   = <permission>
      */

      query = malloc(sizeof("SELECT COUNT(*) FROM permission WHERE username = '") +
                     strlen(cur_trac->user) +
                     sizeof("' AND action = '") +
                     strlen(cur_trac->permission) +
                     sizeof("'") +
                     1);
      if (query == NULL) { error("Out of memory"); }

      strcpy(query, "SELECT COUNT(*) FROM permission WHERE username = '");
      strcat(query, cur_trac->user);
      strcat(query, "' AND action = '");
      strcat(query, cur_trac->permission);
      strcat(query, "'");

#ifdef HAVE_SQLITE3_H
      sqlite3_exec(tracdb, query, &count_callback, NULL, &sql_error);
#else
      sqlite_exec(tracdb, query, &count_callback, NULL, &sql_error);
#endif

      if (sql_error != NULL) {
        fprintf(stderr,
                "Error checking permissions on %s: %s\n",
                db_name,
                sql_error);
        exit(EXIT_FAILURE);
      }

      free(query);

      if (count_result == 0) {

        if (verbose_flag) {
          printf("Adding %s for user %s\n", cur_trac->permission, cur_trac->user);
        }

        query = malloc(sizeof("INSERT INTO permission VALUES ('") +
                       strlen(cur_trac->user) +
                       sizeof("', '") +
                       strlen(cur_trac->permission) +
                       sizeof("')")+100);
        if (query == NULL) { error("Out of memory"); }

        strcpy(query, "INSERT INTO permission VALUES ('");
        strcat(query, cur_trac->user);
        strcat(query, "', '");
        strcat(query, cur_trac->permission);
        strcat(query, "')");

#ifdef HAVE_SQLITE3_H
        sqlite3_exec(tracdb, query, NULL, NULL, &sql_error);
#else
        sqlite_exec(tracdb, query, NULL, NULL, &sql_error);
#endif

        if (sql_error != NULL) {
          fprintf(stderr,
                  "Error inserting permissions on %s: %s\n",
                  db_name,
                  sql_error);
          exit(EXIT_FAILURE);
        }

        free(query);

      }

      cur_trac = cur_trac->next;

    }

    close(tracdb);

  }

#endif

  /* free the symbol table */
  free_groups();

  if (htaccess) {
    if (fclose(htaccess) < 0) {
      error("cannot close htaccess file");
    }
  }

  return 0;
}




