/*! \file svnauthcheck.h
 *  \brief Main header.
 *
 * Main header.
 */

/*
 * File version information:
 * $Id: svnauthcheck.h 1634 2010-05-14 10:05:16Z corti $
 * $Revision: 1634 $
 * $HeadURL: https://svn.id.ethz.ch/bd_webhosting/tools/svnauthcheck/svnauthcheck.h $
 * $Date: 2010-05-14 12:05:16 +0200 (Fri, 14 May 2010) $
 */

#include "config.h"

#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifdef HAVE_SQLITE3_H
#include <sqlite3.h>
#else
#ifdef HAVE_SQLITE_H
#include <sqlite.h>
#endif
#endif

#define MAX_USER_SIZE              32 /*!< Maximum length of a user name (in bytes)       */
#define MAX_REPOSITORY_SIZE        32 /*!< Maximum length of a repository name (in bytes) */
#define MAX_FILE_NAME_LENGTH      512 /*!< Maximum file name length (in bytes)            */
#define MAX_MSG_LEN                64 /*!< Maximum length of an error message             */

#define ALLOW_ENDING_SLASHES_OPT  257 /*!< Value returned by getopt_long when --allow-ending-shlashes is specified */

/* booleans */

#define TRUE  1 /*!< True  */
#define FALSE 0 /*!< False */

/* errors */

#define USAGE_ERROR  1 /*!< Incorrect program usage */
#define SYNTAX_ERROR 2 /*!< Syntax error            */

/*! \def out_of_memory
 */
#define out_of_memory() error("Out of memory")

/* types */

/*!
 * \struct user_type
 * \brief User list member
 *
 * Double linked list to manage users
 */
struct user_type {
  char               name[MAX_USER_SIZE]; /*!< User name                      */
  struct user_type * next;                /*!< Next user                      */
  struct user_type * prev;                /*!< Previous user                  */
  int                unexpanded_group;    /*!< Unknown group not yet expanded */
};

/*!
 * \struct symbol
 * \brief Symbol (user) table entry
 */
struct symbol {
  char              name[MAX_USER_SIZE]; /*!< Symbol (user) name                      */
  struct user_type* list;                /*!< List of users corresponding to the name */
  struct symbol*    next;                /*!< Next symbol                             */
};

/*!
 * \struct child
 * \brief Helper node to link node children
 * \dot
 digraph tree_node {
 node [shape=record, fontname=Helvetica, fontsize=10];
 edge [fontname=Helvetica, fontsize=10];
 n  [ label="node"  URL="\ref node"];
 c1 [ label="child" URL="\ref child"];
 c2 [ label="child" URL="\ref child"];
 c3 [ label="child" URL="\ref child"];
 n1 [ label="node"  URL="\ref node"];
 n2 [ label="node"  URL="\ref node"];
 n3 [ label="node"  URL="\ref node"];
 n  -> c1 [ label = "children" URL="\ref node#children" ];
 c1 -> c2 [ label = "next"     URL="\ref child#next"    ];
 c2 -> c3 [ label = "next"     URL="\ref child#next"    ];
 c1 -> n1 [ label = "node"     URL="\ref child#node"    ];
 c2 -> n2 [ label = "node"     URL="\ref child#node"    ];
 c3 -> n3 [ label = "node"     URL="\ref child#node"    ];
 n1 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 n2 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 n3 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 { rank=same; c1 c2 c3 };
 { rank=same; n1 n2 n3 };
 }
  \enddot
 */
struct child {
  struct node  * node; /*!< node (child) */
  struct child * next; /*!< next child   */
};

/*!
 * \struct node
 * \brief  A node in the directory tree
 * \dot
 digraph tree_node {
 node [shape=record, fontname=Helvetica, fontsize=10];
 edge [fontname=Helvetica, fontsize=10];
 n  [ label="node"  URL="\ref node"  ];
 c1 [ label="child" URL="\ref child" ];
 c2 [ label="child" URL="\ref child" ];
 c3 [ label="child" URL="\ref child" ];
 n1 [ label="node"  URL="\ref node"  ];
 n2 [ label="node"  URL="\ref node"  ];
 n3 [ label="node"  URL="\ref node"  ];
 n  -> c1 [ label = "children" URL="\ref node#children" ];
 c1 -> c2 [ label = "next"     URL="\ref child#next"    ];
 c2 -> c3 [ label = "next"     URL="\ref child#next"    ];
 c1 -> n1 [ label = "node"     URL="\ref child#node"    ];
 c2 -> n2 [ label = "node"     URL="\ref child#node"    ];
 c3 -> n3 [ label = "node"     URL="\ref child#node"    ];
 n1 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 n2 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 n3 -> n  [ label = "parent"   URL="\ref node#parent"   ];
 { rank=same; c1 c2 c3 };
 { rank=same; n1 n2 n3 };
 }
  \enddot
 */
struct node {
  char              * name;     /*!< node name                                                                                       */
  struct user_type  * users;    /*!< users := (readers + parent->readers) - revoked (can be computed only when the tree is complete) */
  struct user_type  * readers;  /*!< readers                                                                                         */
  struct user_type  * revoked;  /*!< users whose read rights are revoked for the current node                                        */
  struct child      * children; /*!< descendants                                                                                     */
  struct node       * parent;   /*!< direct link to the unique parent node                                                           */
};

/*!
 * \struct repository
 * \brief  A subversion repository
 */
struct repository {
  char              * name;     /*!< repository name                                                                                 */
  struct node       * root;     /*!< access tree for the repository                                                                  */
  struct repository * next;     /*!< next repository in the list                                                                     */
};

/*!
 * \struct trac_permission
 * \brief  User/permission mapping for trac
 */
struct trac_permission {
   char                   * user;       /*!< user name         */
   char                   * permission; /*!< permission        */
   struct trac_permission * next;       /*!< next list element */
 };

/*
 * function prototypes
 */

char *                     read_template      (char * name);
extern void                add_node           (const char * repo_name, struct node * n);
extern void                print_node         (struct node * n);
struct node *              dir_search         (struct node * node, char * name);
struct repository *        get_repository     (const char * name);
void                       convert_tree       (struct node * n, char * full_path);
void                       debug_tree         (struct node * n, int level);
void                       error              (const  char * message);
void                       init_dir_tree      (struct repository * repo);
void                       prog_usage         ();
void                       write_location     (FILE * f, char * location, struct node * n);
void                       urldecode          (char * string );
void                       check_ending_slash (char * path);
int                        valid_digit        (char c);
void                       prog_usage         (void);

#ifdef DEBUG
void                       print_list     (struct user_type * list);
#endif

/* group management */
struct user_type *         get_group      (char * name);
struct symbol *            create_group   (char * name, struct user_type * users);
void                       free_groups    (void);

/* user management */
struct user_type *         new_user       (char * name, int unexpanded_group);
int                        contains       (struct user_type * src, struct user_type * usr);
struct user_type *         append         (struct user_type * dest, struct user_type * src);
struct user_type *         clone          (struct user_type * list);
void                       free_users     (struct user_type * list);
void                       print_list     (struct user_type * list);

void check_groups(struct user_type * u);
void check_repository_groups(struct node * n);

